//////////////////////////////////////////////////////////////////////////
//                                                                      //
// log10f.cpp -                                                          //
//                                                                      //
//                              //
//        .                          //
//                                                                      //
// Copyright (c) 2001 RC Module Inc.                                    //
//                                                                      //
//  :  ..                                                 //
//                                                                      //
//   $Revision:: 1     $      $Date:: 2/14/06 8:08p    $                //
//                                                                      //
//////////////////////////////////////////////////////////////////////////

#include "math.h"
#include "auxfd.h"

//  Lg(1+1/2)*c, Lg(1+1/4)*c, Lg(1+1/8)*c, ...
//  c = 040000000
static unsigned int _TLN10[28]={0xb451445,	0x633c60e,	0x346153a,	0x1af5f92,	0xdaf484,
						0x6e51da,	0x375fa8,	0x1bbd9e,	0xde245,	0x6f200,
						0x37937,	0x1bca9,	0xde58,	0x6f2d,	0x3796,	0x1bcb,
						0xde5,	0x6f2,	0x379,	0x1bc,	0xde,	0x6f,	0x37,
						0x1b,	0xd,	0x6,	0x3,	0x1};

//    
//	    lg(x) = lg(~=2)+lg(2^(-1)) = lg(~=2)-lg(2)
static unsigned int _TMLN10[28]={0x7feff0a,	0x76c47b5,	0x72ce3f3,	0x70f420d,	0x700ea02,
					     0x6f9daf5,	0x6f65a87,	0x6f49c11,	0x6f3bd46,	0x6f34dfc,
						 0x6f3165e,	0x6f2fa91,	0x6f2ecab,	0x6f2e5b8,	0x6f2e23e,
						 0x6f2e082,	0x6f2dfa3,	0x6f2df34,	0x6f2defc,	0x6f2dee1,
						 0x6f2ded3,	0x6f2decc,	0x6f2dec8,	0x6f2dec7,	0x6f2dec6,
						 0x6f2dec5,	0x6f2dec5,	0x6f2dec5};

//     10
const unsigned int _TLN210 = 0x13441350;

#ifdef __cplusplus
extern "C" {
#endif

// log10 --
//
//     
//
//    --.
//    CORDIC,    .
float log10f(float f){

	int Mnt;                            //   
    int Exp, Sgn;                       //     
    int Spc;
    int NewExp;                         //   
	unsigned int Arg;                   //     

    //   float
    DestructFloat( f, Mnt, Exp, Sgn, Spc );

	//  
	if ( Exp ==128 || Sgn==1 )	// -x, +inf, NAN:  NAN
		return ConstructFloat( 1, 0x80, 0, 0);

	if ( Mnt == 0 && Exp==-127 )	// 0 -> +inf
		return ConstructFloat( 0, 0x80, 0, 0);

	if (!(Mnt & 0x00800000)){  //  
		Exp++;
		while (!(Mnt & 0x00800000)){
			Mnt <<=1;
			Exp--;
		}
	}
	Mnt <<= 7;
	unsigned int tmp; //       
	int rez = 0;
	Arg= 0x40000000; /*   1,  
								        */
	NewExp = 0;
	int i=0;                        //   
    int j=0;                        //   

	tmp = Arg;

    //      ln(x) = ln(~=2)+ln(2^(-1)) = ln(~=2)-ln(2)
	if (Exp==-1){
		Mnt>>=1;

		while (i<=27){  
			i++;
			tmp -= (tmp>>(i+1)) + ((tmp>>(i))&1); 
			if (tmp>=Mnt){
				if (j==0){
					j=i;
					NewExp=-j+1;
					rez<<=j;
				}
				Arg = tmp;
				rez += _TMLN10[i-1]>>(i-j);

				tmp -= (tmp>>(i+1)) + ((tmp>>(i))&1); 
				if (tmp>=Mnt){
					Arg = tmp;
					rez += _TMLN10[i-1]>>(i-j);
				}
				else
					tmp = Arg;
			}
			else
				tmp = Arg;

		}
		Exp=0;
		Sgn=1;
	}
	else
        //     
		while (i<=27){  
			i++;
			tmp += tmp>>i; 
			if (tmp<=Mnt){
				Arg = tmp;
				rez += _TLN10[i-1];
			}
			else 
				tmp = Arg;
		}
	
	int tln2 = _TLN210;

	if (Exp<-1){
		rez = -rez;
		Exp = -Exp;
		Sgn=1;
	}

	i=0;
	int NewMnt = rez;
    //   :   log(2x) = log2 + log(x)
	while (Exp>0){  
		i++;
		if (Exp & 1){
			NewMnt += tln2;
			rez = NewMnt;
			rez >>=1;
			NewExp = i;
		}
		Exp>>=1;
		NewMnt>>=1;
	}
	if (rez & 0x40000000)
		rez >>=1;
	else
		NewExp--;

    if (rez){
		while (!(rez & 0x20000000)){
			rez <<=1;
			NewExp--;
		}
    }
    else
    {
        NewExp= -127;    
    }

	rez >>=5;	//    , ...
	rez = (rez>>1) + (rez & 1); //   ,  .

	if (rez == 0x01000000){
		rez = 0;
		NewExp++;
	}

    //   float
	return ConstructFloat( rez, NewExp, Sgn, 0);
}

#ifdef __cplusplus
}
#endif
